@@ -4,4 +4,4 @@
 # DeepSpeed Team
 from .config_v2 import RaggedInferenceEngineConfig, DeepSpeedTPConfig
 from .engine_v2 import InferenceEngineV2
-from .engine_factory import build_hf_engine
+from .engine_factory import build_hf_engine, build_engine_from_ds_checkpoint
