@@ -1,15 +1,20 @@
-'''
-Copyright 2022 The Microsoft DeepSpeed Team
-'''
+# Copyright (c) Microsoft Corporation.
+# SPDX-License-Identifier: Apache-2.0
+
+# DeepSpeed Team
+
 import math
 import torch
 from torch.autograd import Function
-from ... import op_builder
 import torch.nn as nn
 from packaging import version as pkg_version
 from deepspeed.utils.logging import log_dist
-# Cuda modules will be imported if needed
-inference_cuda_module = None
+from deepspeed.accelerator import get_accelerator
+from deepspeed.ops.transformer.inference.op_binding.workspace import WorkspaceOp
+from deepspeed.ops.transformer.inference.op_binding.softmax_context import SoftmaxContextOp
+from deepspeed.ops.transformer.inference.op_binding import LinearOp
+from deepspeed.ops.transformer.inference.op_binding.pad_transform import PadTransformOp
+
 minus_inf = -10000.0
 triton_flash_attn = None
 
@@ -28,26 +33,13 @@ def load_triton_flash_attn():
 
 
 class DeepSpeedDiffusersAttentionFunction(Function):
+
     @staticmethod
-    def forward(ctx,
-                input,
-                context,
-                input_mask,
-                config,
-                attn_qkvw,
-                attn_qw,
-                attn_kw,
-                attn_vw,
-                attn_qkvb,
-                num_attention_heads_per_partition,
-                norm_factor,
-                hidden_size_per_partition,
-                attn_ow,
-                attn_ob,
-                do_out_bias,
-                score_context_func,
-                linear_func,
-                triton_flash_attn_kernel):
+    def forward(ctx, input, context, input_mask, config, attn_qkvw, attn_qw, attn_kw, attn_vw, attn_qkvb,
+                num_attention_heads_per_partition, norm_factor, hidden_size_per_partition, attn_ow, attn_ob,
+                do_out_bias, score_context_func, linear_func, pad_transform_func, triton_flash_attn_kernel,
+                rope_theta):
+
         def _transpose_for_context(x):
             x = x.permute(0, 2, 1, 3)
             new_x_layer_shape = x.size()[:-2] + \
@@ -56,32 +48,24 @@ class DeepSpeedDiffusersAttentionFunction(Function):
 
         def _transpose_for_scores(x):
             attention_head_size = x.shape[-1] // num_attention_heads_per_partition
-            new_x_shape = x.size()[:-1] + (num_attention_heads_per_partition,
-                                           attention_head_size)
+            new_x_shape = x.size()[:-1] + (num_attention_heads_per_partition, attention_head_size)
             x = x.reshape(*new_x_shape)
             x = x.permute(0, 2, 1, 3)
             return x.contiguous()
 
         def selfAttention_fp(input, context, input_mask):
-            if config.fp16 and input.dtype == torch.float32:
+            if config.dtype in [torch.half, torch.float16] and input.dtype == torch.float32:
                 input = input.half()
             head_size = input.shape[-1] // config.heads
             do_flash_attn = (head_size <= 128)
             scale = (1 / norm_factor) * (1 / norm_factor)
-            if do_flash_attn and context == None:
-                qkv_out = linear_func(input,
-                                      attn_qkvw,
-                                      attn_qkvb if attn_qkvb is not None else attn_qkvw,
-                                      attn_qkvb is not None,
-                                      do_flash_attn,
-                                      config.heads)
+            if do_flash_attn and context is None:
+                qkv_out = linear_func(input, attn_qkvw, attn_qkvb if attn_qkvb is not None else attn_qkvw, attn_qkvb
+                                      is not None, do_flash_attn, config.heads, False, rope_theta)
 
-                context_layer = triton_flash_attn_kernel(qkv_out[0],
-                                                         qkv_out[1],
-                                                         qkv_out[2],
-                                                         scale,
+                context_layer = triton_flash_attn_kernel(qkv_out[0], qkv_out[1], qkv_out[2], scale,
                                                          input.shape[-2] % 128 == 0)
-                context_layer = _transpose_for_context(context_layer[:,:,:,:head_size])
+                context_layer = _transpose_for_context(context_layer[:, :, :, :head_size])
 
             else:
                 do_flash_attn = False
@@ -95,21 +79,11 @@ class DeepSpeedDiffusersAttentionFunction(Function):
                     query = query.contiguous()
                     key = key.contiguous()
                     value = value.contiguous()
-                query, key, value = inference_cuda_module.pad_transform_fp16(query, key, value, config.heads, do_flash_attn)
-                attention_scores = (torch.matmul(query,
-                                                 key.transpose(-1,
-                                                               -2)) *
-                                    scale).softmax(dim=-1)
-                context_layer = _transpose_for_context(
-                    torch.matmul(attention_scores,
-                                 value))
+                query, key, value = pad_transform_func(query, key, value, config.heads, do_flash_attn)
+                attention_scores = (torch.matmul(query, key.transpose(-1, -2)) * scale).softmax(dim=-1)
+                context_layer = _transpose_for_context(torch.matmul(attention_scores, value))
 
-            output = linear_func(context_layer,
-                                 attn_ow,
-                                 attn_ob,
-                                 do_out_bias,
-                                 False,
-                                 config.heads)
+            output = linear_func(context_layer, attn_ow, attn_ob, do_out_bias, False, config.heads, False, rope_theta)
             return output
 
         output = selfAttention_fp(input, context, input_mask)
@@ -140,15 +114,11 @@ class DeepSpeedDiffusersAttention(nn.Module):
         self.config = config
         self.config.layer_id = DeepSpeedDiffusersAttention.layer_id
         DeepSpeedDiffusersAttention.layer_id += 1
-        device = torch.cuda.current_device() if config.bigscience_bloom else 'cpu'
+        device = get_accelerator().current_device_name() if config.bigscience_bloom else 'cpu'
         qkv_size_per_partition = (self.config.hidden_size // self.config.mp_size) * 3
 
-        data_type = torch.int8 if config.q_int8 else torch.half if config.fp16 else torch.float
-        data_type_fp = torch.half if config.fp16 else torch.float
-        global inference_cuda_module
-        if inference_cuda_module is None:
-            builder = op_builder.InferenceBuilder()
-            inference_cuda_module = builder.load()
+        data_type = self.config.dtype
+        data_type_fp = torch.half if self.config.dtype == torch.int8 else self.config.dtype
 
         if DeepSpeedDiffusersAttention.layer_id == 1:
             log_dist(f"DeepSpeed-Attention config: {self.config.__dict__}", [0])
@@ -173,9 +143,7 @@ class DeepSpeedDiffusersAttention(nn.Module):
                                                 dtype=data_type,
                                                 device=device),
                                     requires_grad=False)
-        self.attn_qkvb = nn.Parameter(torch.empty(qkv_size_per_partition,
-                                                  dtype=data_type_fp,
-                                                  device=device),
+        self.attn_qkvb = nn.Parameter(torch.empty(qkv_size_per_partition, dtype=data_type_fp, device=device),
                                       requires_grad=False)
         out_size_per_partition = self.config.hidden_size // self.config.mp_size
         self.attn_ow = nn.Parameter(torch.empty(out_size_per_partition,
@@ -184,9 +152,7 @@ class DeepSpeedDiffusersAttention(nn.Module):
                                                 device=device),
                                     requires_grad=False)
 
-        self.attn_ob = nn.Parameter(torch.empty(self.config.hidden_size,
-                                                dtype=data_type_fp,
-                                                device=device),
+        self.attn_ob = nn.Parameter(torch.empty(self.config.hidden_size, dtype=data_type_fp, device=device),
                                     requires_grad=False)
         self.do_out_bias = True
 
@@ -197,49 +163,30 @@ class DeepSpeedDiffusersAttention(nn.Module):
         self.hidden_size_per_partition = self.config.hidden_size // self.config.mp_size
         self.hidden_size_per_attention_head = self.config.hidden_size // self.config.heads
 
-        self.norm_factor = math.sqrt(
-            math.sqrt(self.config.hidden_size // self.config.heads))
+        self.norm_factor = math.sqrt(math.sqrt(self.config.hidden_size // self.config.heads))
 
         if self.config.scale_attn_by_inverse_layer_idx is True:
             self.norm_factor *= math.sqrt(self.config.layer_id + 1)
             # https://github.com/huggingface/transformers/blob/v4.24.0/src/transformers/models/gpt2/modeling_gpt2.py#L191
 
-        self.score_context_func = inference_cuda_module.softmax_context_fp32 if (not config.fp16) else \
-                                    inference_cuda_module.softmax_context_fp16
-        self.linear_func = inference_cuda_module.linear_layer_fp16 if config.fp16 else \
-                                    inference_cuda_module.linear_layer_fp32
-        self.allocate_workspace = inference_cuda_module.allocate_workspace_fp32 if not (config.fp16) else \
-                                    inference_cuda_module.allocate_workspace_fp16
+        self.workspace = WorkspaceOp(self.config)
+        self.score_context_func = SoftmaxContextOp(self.config)
+        self.linear_func = LinearOp(self.config)
+        self.pad_transform_func = PadTransformOp(self.config)
+
+    def allocate_workspace(self, size):
+        # Allocate memory only on first layer forward
+        if self.config.layer_id == 0:
+            self.workspace.allocate_workspace(self.config.hidden_size, self.config.heads, size[1], size[0],
+                                              DeepSpeedDiffusersAttention.layer_id, self.config.mp_size, False, 0,
+                                              self.config.max_out_tokens, self.config.min_out_tokens)
 
     def forward(self, input, context=None, input_mask=None):
-        if self.config.layer_id == 0:
-            self.allocate_workspace(self.config.hidden_size,
-                                    self.config.heads,
-                                    input.size()[1],
-                                    input.size()[0],
-                                    DeepSpeedDiffusersAttention.layer_id,
-                                    self.config.mp_size,
-                                    False,
-                                    0,
-                                    self.config.max_out_tokens)
+        self.allocate_workspace(input.size())
         output = DeepSpeedDiffusersAttentionFunction.apply(
-            input,
-            context,
-            input_mask,
-            self.config,
-            self.attn_qkvw,
-            self.attn_qw,
-            self.attn_kw,
-            self.attn_vw,
-            self.attn_qkvb,
-            self.num_attention_heads_per_partition,
-            self.norm_factor,
-            self.hidden_size_per_partition,
-            self.attn_ow,
-            self.attn_ob,
-            self.do_out_bias,
-            self.score_context_func,
-            self.linear_func,
-            self.triton_flash_attn_kernel)
+            input, context, input_mask, self.config, self.attn_qkvw, self.attn_qw, self.attn_kw, self.attn_vw,
+            self.attn_qkvb, self.num_attention_heads_per_partition, self.norm_factor, self.hidden_size_per_partition,
+            self.attn_ow, self.attn_ob, self.do_out_bias, self.score_context_func, self.linear_func,
+            self.pad_transform_func, self.triton_flash_attn_kernel, self.config.rope_theta)
 
         return output
