@@ -1,7 +1,7 @@
-"""
-"Copyright 2020 The Microsoft DeepSpeed Team.
-Licensed under the MIT license.
-"""
+# Copyright (c) Microsoft Corporation.
+# SPDX-License-Identifier: Apache-2.0
+
+# DeepSpeed Team
 
 from .partition_parameters import ZeroParamType
 from .partition_parameters import ZeroParamStatus
@@ -11,3 +11,7 @@ from .partition_parameters import register_external_parameter
 
 from .tiling import TiledLinear
 from .tiling import TiledLinearReturnBias
+
+from .mics import MiCS_Init
+
+from .stage3 import unwrap_model_for_generation
